#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Analysis;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic.Regression {
  /// <summary>
  /// An operator that analyzes the training best symbolic regression solution for multi objective symbolic regression problems.
  /// </summary>
  [Item("SymbolicRegressionMultiObjectiveTrainingBestSolutionAnalyzer", "An operator that analyzes the training best symbolic regression solution for multi objective symbolic regression problems.")]
  [StorableClass]
  public sealed class SymbolicRegressionMultiObjectiveTrainingBestSolutionAnalyzer : SymbolicDataAnalysisMultiObjectiveTrainingBestSolutionAnalyzer<ISymbolicRegressionSolution>,
    ISymbolicDataAnalysisInterpreterOperator, ISymbolicDataAnalysisBoundedOperator {
    private const string ProblemDataParameterName = "ProblemData";
    private const string SymbolicDataAnalysisTreeInterpreterParameterName = "SymbolicDataAnalysisTreeInterpreter";
    private const string EstimationLimitsParameterName = "EstimationLimits";
    private const string MaximumSymbolicExpressionTreeLengthParameterName = "MaximumSymbolicExpressionTreeLength";

    #region parameter properties
    public ILookupParameter<IRegressionProblemData> ProblemDataParameter {
      get { return (ILookupParameter<IRegressionProblemData>)Parameters[ProblemDataParameterName]; }
    }
    public ILookupParameter<ISymbolicDataAnalysisExpressionTreeInterpreter> SymbolicDataAnalysisTreeInterpreterParameter {
      get { return (ILookupParameter<ISymbolicDataAnalysisExpressionTreeInterpreter>)Parameters[SymbolicDataAnalysisTreeInterpreterParameterName]; }
    }
    public IValueLookupParameter<DoubleLimit> EstimationLimitsParameter {
      get { return (IValueLookupParameter<DoubleLimit>)Parameters[EstimationLimitsParameterName]; }
    }
    public ILookupParameter<IntValue> MaximumSymbolicExpressionTreeLengthParameter {
      get { return (ILookupParameter<IntValue>)Parameters[MaximumSymbolicExpressionTreeLengthParameterName]; }
    }
    #endregion

    [StorableConstructor]
    private SymbolicRegressionMultiObjectiveTrainingBestSolutionAnalyzer(bool deserializing) : base(deserializing) { }
    private SymbolicRegressionMultiObjectiveTrainingBestSolutionAnalyzer(SymbolicRegressionMultiObjectiveTrainingBestSolutionAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public SymbolicRegressionMultiObjectiveTrainingBestSolutionAnalyzer()
      : base() {
      Parameters.Add(new LookupParameter<IRegressionProblemData>(ProblemDataParameterName, "The problem data for the symbolic regression solution."));
      Parameters.Add(new LookupParameter<ISymbolicDataAnalysisExpressionTreeInterpreter>(SymbolicDataAnalysisTreeInterpreterParameterName, "The symbolic data analysis tree interpreter for the symbolic expression tree."));
      Parameters.Add(new ValueLookupParameter<DoubleLimit>(EstimationLimitsParameterName, "The lower and upper limit for the estimated values produced by the symbolic regression model."));
      Parameters.Add(new LookupParameter<IntValue>(MaximumSymbolicExpressionTreeLengthParameterName, "Maximal length of the symbolic expression."));
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      if (!Parameters.ContainsKey(MaximumSymbolicExpressionTreeLengthParameterName))
        Parameters.Add(new LookupParameter<IntValue>(MaximumSymbolicExpressionTreeLengthParameterName, "Maximal length of the symbolic expression."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new SymbolicRegressionMultiObjectiveTrainingBestSolutionAnalyzer(this, cloner);
    }

    protected override ISymbolicRegressionSolution CreateSolution(ISymbolicExpressionTree bestTree, double[] bestQuality) {
      var model = new SymbolicRegressionModel((ISymbolicExpressionTree)bestTree.Clone(), SymbolicDataAnalysisTreeInterpreterParameter.ActualValue, EstimationLimitsParameter.ActualValue.Lower, EstimationLimitsParameter.ActualValue.Upper);
      if (ApplyLinearScalingParameter.ActualValue.Value) model.Scale(ProblemDataParameter.ActualValue);
      return new SymbolicRegressionSolution(model, (IRegressionProblemData)ProblemDataParameter.ActualValue.Clone());
    }

    public override IOperation Apply() {
      var operation = base.Apply();

      var paretoFront = TrainingBestSolutionsParameter.ActualValue;

      IResult result;
      ScatterPlot qualityToTreeSize;
      if (!ResultCollection.TryGetValue("Pareto Front Analysis", out result)) {
        qualityToTreeSize = new ScatterPlot("Quality vs Tree Size", "");
        qualityToTreeSize.VisualProperties.XAxisMinimumAuto = false;
        qualityToTreeSize.VisualProperties.XAxisMaximumAuto = false;
        qualityToTreeSize.VisualProperties.YAxisMinimumAuto = false;
        qualityToTreeSize.VisualProperties.YAxisMaximumAuto = false;

        qualityToTreeSize.VisualProperties.XAxisMinimumFixedValue = 0;
        qualityToTreeSize.VisualProperties.XAxisMaximumFixedValue = MaximumSymbolicExpressionTreeLengthParameter.ActualValue.Value;
        qualityToTreeSize.VisualProperties.YAxisMinimumFixedValue = 0;
        qualityToTreeSize.VisualProperties.YAxisMaximumFixedValue = 2;
        ResultCollection.Add(new Result("Pareto Front Analysis", qualityToTreeSize));
      } else {
        qualityToTreeSize = (ScatterPlot)result.Value;
      }

      var problemData = ProblemDataParameter.ActualValue;
      var constantValue = problemData.Dataset.GetDoubleValues(problemData.TargetVariable, problemData.TrainingIndices).Average();
      var constantSolution = CreateConstantSymbolicRegressionSolution(constantValue);

      int previousTreeLength = constantSolution.Model.SymbolicExpressionTree.Length;
      var sizeParetoFront = new LinkedList<ISymbolicRegressionSolution>();
      sizeParetoFront.AddLast(constantSolution);

      foreach (var solution in paretoFront.OrderBy(s => s.Model.SymbolicExpressionTree.Length)) {
        int treeLength = solution.Model.SymbolicExpressionTree.Length;
        if (!sizeParetoFront.Any()) sizeParetoFront.AddLast(solution);
        if (solution.TrainingNormalizedMeanSquaredError < sizeParetoFront.Last.Value.TrainingNormalizedMeanSquaredError) {
          if (treeLength == previousTreeLength)
            sizeParetoFront.RemoveLast();
          sizeParetoFront.AddLast(solution);
        }
        previousTreeLength = treeLength;
      }

      qualityToTreeSize.Rows.Clear();
      var trainingRow = new ScatterPlotDataRow("Training NMSE", "", sizeParetoFront.Select(x => new Point2D<double>(x.Model.SymbolicExpressionTree.Length, x.TrainingNormalizedMeanSquaredError)));
      trainingRow.VisualProperties.PointSize = 5;
      var testRow = new ScatterPlotDataRow("Test NMSE", "", sizeParetoFront.Select(x => new Point2D<double>(x.Model.SymbolicExpressionTree.Length, x.TestNormalizedMeanSquaredError)));
      testRow.VisualProperties.PointSize = 5;
      qualityToTreeSize.Rows.Add(trainingRow);
      qualityToTreeSize.Rows.Add(testRow);

      double trainingHyperVolume, testHyperVolume;
      CalculateNormalizedHyperVolume(sizeParetoFront, out trainingHyperVolume, out testHyperVolume);

      DoubleValue trainingHyperVolumeResult, testHyperVolumeResult;
      if (!ResultCollection.TryGetValue("HyperVolume training", out result)) {
        trainingHyperVolumeResult = new DoubleValue();
        ResultCollection.Add(new Result("HyperVolume training", trainingHyperVolumeResult));
      } else trainingHyperVolumeResult = (DoubleValue)result.Value;

      if (!ResultCollection.TryGetValue("HyperVolume test", out result)) {
        testHyperVolumeResult = new DoubleValue();
        ResultCollection.Add(new Result("HyperVolume test", testHyperVolumeResult));
      } else testHyperVolumeResult = (DoubleValue)result.Value;

      trainingHyperVolumeResult.Value = trainingHyperVolume;
      testHyperVolumeResult.Value = testHyperVolume;

      return operation;
    }

    private void CalculateNormalizedHyperVolume(IEnumerable<ISymbolicRegressionSolution> solutions, out double trainingHyperVolume, out double testHyperVolume) {
      trainingHyperVolume = 0.0;
      testHyperVolume = 0.0;

      var prevX = 0;
      var prevYTraining = 1.0;
      var prevYTest = 1.0;
      var treeSize = 0;

      foreach (var solution in solutions) {
        treeSize = solution.Model.SymbolicExpressionTree.Length;
        trainingHyperVolume += (prevYTraining + solution.TrainingNormalizedMeanSquaredError) * (treeSize - prevX) / 2;
        testHyperVolume += (prevYTest + solution.TestNormalizedMeanSquaredError) * (treeSize - prevX) / 2;
        prevX = treeSize;
        prevYTraining = solution.TrainingNormalizedMeanSquaredError;
        prevYTest = solution.TestNormalizedMeanSquaredError;
      }

      trainingHyperVolume = trainingHyperVolume / treeSize;
      testHyperVolume = testHyperVolume / treeSize;
    }

    private ISymbolicRegressionSolution CreateConstantSymbolicRegressionSolution(double constantValue) {
      var grammar = new ArithmeticExpressionGrammar();
      var rootNode = (SymbolicExpressionTreeTopLevelNode)grammar.ProgramRootSymbol.CreateTreeNode();
      var startNode = (SymbolicExpressionTreeTopLevelNode)grammar.StartSymbol.CreateTreeNode();
      var constantTreeNode = new ConstantTreeNode(new Constant()) { Value = constantValue };
      rootNode.AddSubtree(startNode);
      startNode.AddSubtree(constantTreeNode);
      var tree = new SymbolicExpressionTree();
      tree.Root = rootNode;

      var model = new SymbolicRegressionModel(tree, SymbolicDataAnalysisTreeInterpreterParameter.ActualValue);
      var solution = new SymbolicRegressionSolution(model, ProblemDataParameter.ActualValue);
      return solution;
    }
  }
}
